#include <math.h>
#include <stdlib.h>
#include "lis.h"
#include "lis_solver.h"

int lis_gmres(LIS_VECTOR b, LIS_VECTOR x, LIS_VECTOR work[],
		   LIS_SCALAR params[], int options[],
		   LIS_MATRIX A, LIS_PRECON M, LIS_SCALAR status[])
{
	LIS_VECTOR r,s, z, *v;
	LIS_SCALAR *h;
	LIS_SCALAR aa,bb,rr,r0,scale,a2,b2,t;
	LIS_SCALAR bnrm2, nrm2, rnorm;
	LIS_SCALAR tol;
	int iter,maxiter,n,np,i,j,k,m,output;
	int ii,i1,iiv,i1v,iih,i1h,jj;
	int w_dim, h_dim;
	int cs,sn;
	double times,ptimes;

	n       = A->n;
	np      = A->np + options[LIS_OPTIONS_PADDING]; 
	w_dim   = np;
	maxiter = options[LIS_OPTIONS_MAXITER];
	output  = options[LIS_OPTIONS_OUTPUT];
	m       = options[LIS_OPTIONS_RESTART];
	h_dim   = m+1;
	tol     = params[LIS_PARAMS_RESID];

	r       = work[0];
	s       = work[1];
	z       = work[2];
	v       = &work[3];

	h       = (LIS_SCALAR *)lis_malloc( sizeof(LIS_SCALAR) * (h_dim+1) * (h_dim+2) );
	cs      = (m+1)*h_dim;
	sn      = (m+2)*h_dim;

	ptimes  = 0.0;

	// Initial Residual
	/* ||b|| */
	lis_vector_nrm2(b,&bnrm2);
	if( bnrm2 == 0.0 )
	{
		bnrm2 = 1.0;
	}
	else
	{
		bnrm2 = 1.0 / bnrm2;
	}
	lis_vector_dot(x,x,&nrm2);
	if( nrm2 != 0.0 )
	{
		lis_matvec(A,x,v[0]); /* v = b - Ax */
		lis_vector_xpay(b,-1,v[0]);
		lis_vector_nrm2(v[0],&nrm2);
		nrm2 = nrm2 * bnrm2;
		if( nrm2 <= tol )
		{
			lis_free(h);
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_SUCCESS;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)1;
			status[LIS_STATUS_RESID] = nrm2;
			return LIS_SUCCESS;
		}
	}
	else
	{
		lis_vector_copy(b,v[0]);
	}

	iter=0;
	while( iter<=maxiter )
	{
		/* first column of V */
		/* v = r / ||r||_2 */
		lis_vector_nrm2(v[0],&rnorm);
		lis_vector_scale(1.0/rnorm,v[0]);

		/* s = ||r||_2 e_1 */
		lis_vector_set_all(0,s);
		s->value[0] = rnorm;

		i = 0;
		do
		{
			iter++;
			i++;
			ii  = i-1;
			i1  = i;
			iiv = i-1;
			i1v = i;
//			iiv = (i-1)*w_dim;
//			i1v = i*w_dim;
			iih = (i-1)*h_dim;
			i1h = i*h_dim;


			/* z = M^-1 v */
			times = lis_wtime();
			lis_psolve(M, v[iiv], z);
			ptimes += lis_wtime()-times;

			/* v = Az */
			lis_matvec(A,z, v[i1v]);

			for(k=0;k<i;k++)
			{
				/* h[k,i]   = <w,v[k]>       */
				/* w        = w - h[k,i]v[k] */
				lis_vector_dot(v[i1v],v[k],&t);
				h[k + iih] = t;
				lis_vector_axpy(-t,v[k],v[i1v]);
			}
			/* h[i+1,i] = ||w||          */
			/* v[i+1]   = w / h[i+1,i]   */
			lis_vector_nrm2(v[i1v],&t);
			h[i1 + iih] = t;
			lis_vector_scale(1.0/t,v[i1v]);

			for(k=1;k<=ii;k++)
			{
				jj  = k-1;
				t   =  h[jj + iih];
				aa  =  h[jj + cs]*t;
				aa +=  h[jj + sn]*h[k  + iih];
				bb  = -h[jj + sn]*t;
				bb +=  h[jj + cs]*h[k  + iih];
				h[jj + iih] = aa;
				h[k  + iih] = bb;
			}
			aa = h[ii + iih];
			bb = h[i1 + iih];
			a2 = aa*aa;
			b2 = bb*bb;
			rr = sqrt(a2 + b2);
			if( rr==0.0 ) rr=1.0e-17;
			h[ii + cs] = aa / rr;
			h[ii + sn] = bb / rr;
			s->value[i1] = -h[ii + sn]*s->value[ii];
			s->value[ii] =  h[ii + cs]*s->value[ii];

			aa  =  h[ii + cs]*h[ii + iih];
			aa +=  h[ii + sn]*h[i1 + iih];
			h[ii   + iih] = aa;

			/* convergence check */
			nrm2 = fabs(s->value[i1]) * bnrm2;

			if( output )
			{
			if( output & LIS_PRINT_MEM ) lis_set_residual_history(nrm2);
			if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
			}

			if( tol >= nrm2 ) break;
		} while( i<m && iter <=maxiter );

//		update(m-1,n,x,h,h_dim,y,s,v,w_dim);
		/* Solve H*Y =S for upper triangular H */
		s->value[ii] = s->value[ii] / h[ii + iih];
		for(k=1;k<=ii;k++)
		{
			jj = ii-k;
			t  = s->value[jj];
			for(j=jj+1;j<=ii;j++)
			{
				t -= h[jj + j*h_dim]*s->value[j];
			}
			s->value[jj] = t / h[jj + jj*h_dim];
		}
		/*
		for(k=ii-1;k>=0;k--)
		{
			t = s[k];
			for(j=k+1;j<=ii;j++)
			{
				t -= h[k + j*h_dim]*s[j];
			}
			s[k] = t / h[k + k*h_dim];
		}*/
		/* x = x + yv */
		for(k=0;k<n;k++)
		{
			z->value[k] = s->value[0]*v[0]->value[k];
		}
		for(j=1;j<=ii;j++)
		{
			lis_vector_axpy(s->value[j],v[j],z);
		}
		/* r = M^-1 z */
		times = lis_wtime();
		lis_psolve(M, z, r);
		ptimes += lis_wtime()-times;

		/* x = x + r */
		lis_vector_axpy(1,r,x);

		if( tol >= nrm2 )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_SUCCESS;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			status[LIS_STATUS_PTIMES] = ptimes;
			lis_free(h);
			return LIS_SUCCESS;
		}

		for(j=1;j<=i;j++)
		{
			jj = i1-j+1;
			s->value[jj-1] = -h[jj-1 + sn] * s->value[jj];
			s->value[jj]   =  h[jj-1 + cs] * s->value[jj];
		}

		for(j=0;j<=i1;j++)
		{
			t = s->value[j];
			if( j==0 ) t = t-1.0;
			lis_vector_axpy(t,v[j],v[0]);
		}
	}

	lis_free(h);
	status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_MAXITER;
	status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
	status[LIS_STATUS_RESID] = nrm2;
	return LIS_MAXITER;
}

